from wand.image import Image
from wand.color import Color
from wand.drawing import Drawing
from wand.font import Font
from datetime import datetime, timedelta
import math
from types import SimpleNamespace
import logging
import tempfile
import subprocess
import hashlib

from . import timer
from .collect_tk import Worker
from .res import CHAR_TO_FILE, colours, colr_res, file_res
from . import game as gamelib

logger = logging.getLogger(__name__)

class GameImage:
    handlers = dict()

    def __init__(self, font=SimpleNamespace(family='Noto-Sans-CJK-TC', size=18)):
        self.font = font

    @classmethod
    def handler(cls, game):
        def wrap(fn):
            cls.handlers[game] = fn
            return fn
        return wrap

    def populate_result_grid(self, game, result_grid):
        return self.handlers.get(game.game_name, self._populate_result_grid)(self.font, game, result_grid)

    def _populate_result_grid(self, _, game, result_grid):
        for row in game.result_rows:
            if row == '':
                for i in range(game.cols):
                    with Image(width=20, height=3, filename=colr_res(colours.white)) as item:
                        result_grid.image_add(item)
            else:
                for char in row:
                    if resource := CHAR_TO_FILE.get(char):
                        with Image(width=20, height=20, filename=resource) as item:
                            result_grid.image_add(item)
        result_grid.montage(tile=f'{game.cols}x{game.rows}', thumbnail='+2+2')
        return result_grid

    def make_image(self, game, game_image=None):
        if game_image is None:
            game_image = Image()
        with Image() as result_grid:
            game_image.font = Font(
                self.font.family,
                self.font.size,
                colours.black if game.success else colours.red
            )
            game_image.read(filename=f"label:{game.caption}")
            game_image.image_add(self.populate_result_grid(game, result_grid))
            game_image.smush(stacked=True, offset=5)
            logger.info(f'Constructed image for {game}')
            return game_image

@GameImage.handler(gamelib.MISSING_GAME)
def missing_image_result_grid(_, _game, result_grid):
    result_grid.read(filename=file_res('missing-image.png'))
    return result_grid

@GameImage.handler('diffle')
def diffle_result_grid(_, game, result_grid):
    if isinstance(game.image_data, str):
        result_grid.read(filename=game.image_data)
    elif isinstance(game.image_data, bytes):
        result_grid.read(blob=game.image_data)
    result_grid.chop(width=0, height=70, gravity='north')
    return result_grid

@GameImage.handler('redactle')
def redactle_result_grid(font, game, result_grid):
    result_grid.read(filename=file_res('redactle-base.png'))
    guess_width = math.log(game.extra_results.guesses, 10) * 100
    accuracy_width = game.extra_results.accuracy / 100 * 300
    with Drawing() as ctx:
        ctx.fill_color = Color(colours.red)
        ctx.rectangle(left=80, top=20, width=guess_width, height=10)
        ctx.fill_color = Color(colours.green)
        ctx.rectangle(left=80, top=60, width=accuracy_width, height=10)
        ctx.font = font.family
        ctx.font_size = font.size
        ctx.fill_color = Color('black')
        ctx.text(0, 18, str(game.extra_results.guesses))
        ctx.text(0, 55, str(game.extra_results.accuracy) + '%')
        ctx(result_grid)
    return result_grid

class Timeline:
    minutes_per_row = 30
    def __init__(self, config):
        self.config = config
        self.outcfg = config.outimg_config
        self.timeline_spec = config.timeline

    def _halfhour_bucket(self, timestamp):
        return timestamp.replace(minute=((timestamp.minute // self.minutes_per_row)
                                         * self.minutes_per_row),
                                 second=0,
                                 microsecond=0)

    def _read_timestamp(self, timestring):
        return datetime.fromisoformat(timestring).astimezone(self.config.timezone)

    def _in_bucket(self, timestamp, buckets):
        halfhour = self._halfhour_bucket(timestamp)
        for i, ts in enumerate(buckets):
            if halfhour == ts:
                return i

    def _timeline_rectangle(self, event, buckets, width_per_second, height_per_bucket, left_edge):
        event_timestamp = self._read_timestamp(event['timestamp'])
        x = (left_edge
             + (event_timestamp.minute % self.minutes_per_row * 60 + event_timestamp.second)
             * width_per_second)
        y = self._in_bucket(event_timestamp, buckets) * height_per_bucket
        w = event['duration'] * width_per_second
        h = height_per_bucket
        return (x, y, w, h)

    def make_image(self, events):
        width_per_second = self.timeline_spec.width_per_second
        height_per_bucket = self.timeline_spec.height_per_bucket
        left_edge = self.timeline_spec.left_edge
        border_width = self.timeline_spec.border_width
        buckets = list(sorted({self._halfhour_bucket(self._read_timestamp(i['timestamp']))
                               for i in events}))

        # If there are no events, then don't bother with a timeline:
        if len(events) == 0:
            return Image(width=self.outcfg.width, height=border_width,
                         pseudo=colr_res(colours.border))
        # If the last entry extends past the half-hour mark, then add a new bucket:
        last_entry = events[-1]
        last_entry_ts = self._read_timestamp(last_entry['timestamp'])
        if (self._halfhour_bucket(last_entry_ts) !=
            self._halfhour_bucket(last_entry_ts
                                  + timedelta(seconds=last_entry['duration']))):
            buckets.append(buckets[-1] + timedelta(minutes=self.minutes_per_row))

        img = Image(width=int(width_per_second * self.minutes_per_row * 60) + left_edge,
                    height=len(buckets) * height_per_bucket,
                    pseudo=colr_res(colours.border))
        img.background_color = colours.border
        for i, b in enumerate(buckets): # Half hour labels
            img.caption(b.strftime('%H:%M'),
                        left=0, top=i * height_per_bucket,
                        width=left_edge, height=height_per_bucket,
                        gravity='center',
                        font=Font(self.timeline_spec.font,
                                  self.timeline_spec.font_size, colours.white))

        with Drawing() as ctx:
            # Background
            ctx.fill_color = Color(colours.white)
            ctx.rectangle(left=left_edge, top=0,
                          width=width_per_second * self.minutes_per_row * 60,
                          height=len(buckets) * height_per_bucket)
            for i in range(1, self.minutes_per_row): # 1 minute marks
                ctx.fill_color = Color(colours.border if i % 5 == 0
                                       else self.timeline_spec.colour)
                ctx.line((i * 60 * width_per_second + left_edge, 0),
                         (i * 60 * width_per_second + left_edge, height_per_bucket * len(buckets)))
            for i in events:
                ctx.fill_color = Color('#' + hashlib.md5(i['data']['title'].encode('utf-8')).hexdigest()[0:6])
                x, y, w, h = self._timeline_rectangle(i, buckets, width_per_second, height_per_bucket, left_edge)
                ctx.rectangle(left=x, top=y, width=w, height=h)
                if x + w > img.width: # Wrap around if event crosses the half-hour mark
                    ctx.rectangle(left=left_edge, top=y + height_per_bucket,
                                  right=x + w - width_per_second * self.minutes_per_row * 60,
                                  height=h)
            ctx.draw(img)
        img.extent(width=self.outcfg.width, height=img.height + 2 * border_width,
                   gravity='center')
        return img

class Header:
    def __init__(self, config):
        self.outcfg = config.outimg_config
        self.date = config.date

    def make_image(self, time_taken):
        img = Image()
        with Image(pseudo=colr_res(colours.theme2), height=45, width=self.outcfg.width) as title:
            title.font = Font(self.outcfg.font, self.outcfg.title_size, colours.theme1)
            title.label("Wordle-Like Statistics", gravity='center')
            img.image_add(title)
        with Image(pseudo=colr_res(colours.theme2), height=40, width=self.outcfg.width) as subtitle:
            subtitle.font = Font(self.outcfg.font, self.outcfg.subtitle_size, colours.theme1)
            subtitle.label(
                "{date:%Y-%m-%d}{time_taken}".format(
                    date=self.date,
                    time_taken=f' - {time_taken} minutes' if time_taken else '',
                ),
                gravity='center',
            )
            img.image_add(subtitle)
        return img

class Mkimg:
    def __init__(self, config):
        self.config = config
        self.layout = config.dat['layouts'][0]
        self.outcfg = config.outimg_config
        self.activitywatch = timer.Query(config)
        self.header = Header(config)
        self.timeline = Timeline(config)
        self.game_image_maker = GameImage(config.outimg_font)
        self.game_parser = gamelib.GameParser(config.parsed_games)
        self.collection = self.config.collection.collection

    def load_games(self):
        saved = dict()
        for game in self.collection:
            saved[game.short_name] = self.game_parser.load_record(game.result)
        return saved

    def summary_subsequence(self, seq, layout, keys, vertical, recur=0):
        if isinstance(layout, str):
            with Image() as game_image:
                game = keys.get(layout)
                if game is None:
                    logger.warn(f'Game {layout} not found in collected games; '
                                'using fallback missing game picture')
                    game = gamelib.ParsedGame.missing_game(layout)
                seq.image_add(
                    self.game_image_maker.make_image(game, game_image))
        elif isinstance(layout, list):
            with Image() as subseq:
                for i in layout:
                    self.summary_subsequence(subseq, i, keys, not vertical, recur + 1)
                if recur == 0: # Centre the highest-level rows
                    for i in range(subseq.iterator_length()):
                        subseq.iterator_set(i)
                        logger.info(f'Row {i} is {subseq.width} x {subseq.height}')
                        subseq.extent(width=self.outcfg.width, gravity='north')
                subseq.smush(stacked=vertical, offset=self.outcfg.game_separation)
                seq.image_add(subseq)

    def summary_image(self):
        with (self.header.make_image(self.activitywatch.time_taken()) as final,
              self.timeline.make_image(self.activitywatch.matching_events()) as timeline,
              tempfile.NamedTemporaryFile(prefix=self.config.program_name.lower()) as f):
            final.gravity = 'north'
            final.image_add(timeline)
            self.summary_subsequence(final, self.layout, self.load_games(), True)
            final.smush(stacked=True,)
            final.format = 'png'
            final.save(f)
            outfile = self.config.autowordle_outfile()
            subprocess.run(['pngcrush', f.name, outfile])
            logger.info(f'Image sent to {outfile}.')

def send_result_to_clipboard(cfg):
    subprocess.run(
        ['xclip', '-sel', 'c', '-target', 'text/uri-list'],
        input=f'file://{cfg.autowordle_outfile()}',
        text=True)

class MkimgWorker(Worker):
    event_name = "<<MkimgDone>>"
    def exec(self):
        Mkimg(self.config).summary_image()
        send_result_to_clipboard(self.config)
    def hook(self):
        self.gui.new_status(f'Image constructed and sent to {self.config.autowordle_outfile()}')



#!/usr/bin/env python3

from datetime import datetime, timedelta
from aw_client import ActivityWatchClient
from pathlib import Path
import logging

QUERY = """
events = query_bucket('aw-watcher-window_fiventsu');
not_afk = flood(query_bucket(find_bucket("aw-watcher-afk_fiventsu")));
not_afk = filter_keyvals(not_afk, "status", ["not-afk"]);
RETURN = filter_period_intersect(events, not_afk);
"""
_logger = logging.getLogger(__name__)

class Query:
    def __repr__(self):
        return "<{class_name} {has_data} starting {date:%Y-%m-%d %H:%M:%S}>".format(
            class_name=type(self).__qualname__,
            has_data="NO DATA" if self.events is None else "loaded",
            date=self.date
        )
    def __init__(self, cfg):
        self.events = None
        self.timezone = cfg.timezone
        last_completed_time = datetime.fromtimestamp(
            max(f.stat().st_mtime
                for f
                in Path(cfg.autowordle_outfile()).parent.glob('*.png')),
            tz=self.timezone)
        self.date = max(cfg.date - timedelta(hours=18), last_completed_time)
        _logger.info(f'Timer is counting from {self.date:%d%H%MJ}')
        self.window_titles = cfg.window_titles()

    def _ensure_query_run(self):
        if self.events is None:
            cl = ActivityWatchClient("wordle-clone-query-" + __name__)
            self.events = cl.query(QUERY, [(self.date, self.date + timedelta(days=1))])[0]

    def time_taken(self):
        self._ensure_query_run()
        return int(round(sum(e["duration"] for e in self.matching_events()) / 60, 0))

    def matching_events(self):
        self._ensure_query_run()
        return [event for event in self.events
                if any(event["data"]["title"].startswith(keys)
                       for keys in self.window_titles)]



from tkinter import PhotoImage, NORMAL, END, DISABLED, Text, HORIZONTAL, ttk
from ttkthemes import ThemedTk
import threading
import subprocess
import logging

from .collect import GameStatus, CollectionStatus, clipboard
from .res import file_res

logger = logging.getLogger(__name__)

class TetsudoruCommands:
    def __init__(self, gui=None):
        self.gui = gui

    strings = [
        "tetsudoru -1 | grep -v '[きぎただあさざかが]'",
        "tetsudoru -1 | grep -v '[きぎただあさざかがつっづしじんまちぢ]'",
        "tetsudoru -1 | grep -v '[きぎただあさざかがつっづしじんまちぢいよひびぴらの]'",
    ]
    def to_clipboard(self, idx):
        subprocess.run(['xclip', '-sel', 'c'], input=self.strings[idx - 1], text=True)
        if self.gui is not None:
            self.gui.new_status(f"Sent search string for Tetsudoru ({idx} guesses)")

class ClipwaitThread:
    event_name = "<<NewClipboard>>"
    finish_event_name = "<<CollectionComplete>>"
    def __init__(self, gui, *, daemon=False):
        self.gui = gui
        self.col = gui.col
        self.daemon = daemon
        gui.root.bind(self.event_name, lambda e: self.hook())
    def _exec(self):
        subprocess.run(["clipnotify"])
        self.gui.root.event_generate(self.event_name)
    def start(self):
        threading.Thread(target=self._exec, daemon=self.daemon).start()
        logger.debug('New clip wait thread spawned')
    def hook(self):
        dat = clipboard()
        # self.gui.update_clipboard(dat)
        if matchp := self.col.find_matching_game(dat):
            self.gui.update_games_view(matchp)
        match self.col.status():
            case CollectionStatus.INCOMPLETE:
                self.start()
            case CollectionStatus.COMPLETE_FOR_NOW:
                self.gui.new_status('Collection complete; no longer listening on clipboard.')
            case CollectionStatus.COMPLETE:
                self.gui.root.event_generate(self.finish_event_name)

class Worker:
    event_name = None
    def __init__(self, config):
        self.config = config
        self.gui = None
    def inject_gui(self, gui, kickoff_event):
        self.gui = gui
        gui.root.bind(kickoff_event, lambda e: self.start(), add='+')
        gui.root.bind(self.event_name, lambda e: self.hook())
    def exec(self):
        pass
    def _exec(self):
        self.exec()
        if self.gui is not None:
            self.gui.root.event_generate(self.event_name)
    def start(self):
        self.thread = threading.Thread(target=self._exec)
        self.thread.start()
    def hook(self):
        pass

class GUI:
    _icons = {
        GameStatus.COLLECTED: 'emblem-default-symbolic.png',
        GameStatus.UNCOLLECTED: 'emblem-error.png',
        GameStatus.UNELIGIBLE: 'emblem-readonly.png',
        GameStatus.PARTIALLY_COLLECTED: 'emblem-remove.png',
        GameStatus.OPTIONAL: 'emblem-question.png',
    }

    def __init__(self, config):
        self.col = config.collection
        self.root = ThemedTk(theme='breeze', className=config.program_name.lower())
        self.workers = list()
        self.date = config.date
        self.complete = False
        self.tetsudoru = TetsudoruCommands(self)
        self.root.title(f'{config.program_name}: {config.date:%Y-%m-%d}')
        self.root.grid_rowconfigure(0, weight=1)
        self.root.grid_columnconfigure(0, weight=1)
        for (k, v) in self._icons.items():
            self._icons[k] = PhotoImage(file=file_res(v), format='png')
        self.clipwait_thread = ClipwaitThread(self, daemon=True)
        self.col.inject_gui(self)

    def _title(self):
        return f"Collecting for {self.date:%Y-%m-%d}"

    def add_worker(self, worker, kickoff_event):
        worker.inject_gui(self, kickoff_event)
        self.workers.append(worker)
        logger.info(f'Added worker {worker} on event {kickoff_event}')

    def update_clipboard(self, new_text):
        self.clipboard_display.config(state=NORMAL)
        self.clipboard_display.delete('1.0', END)
        if isinstance(new_text, str):
            self.clipboard_display.insert(END, new_text)
        else:
            self.clipboard_display.insert(END, '<Non-text data>')
        self.clipboard_display.config(state=DISABLED)

    def update_games_view(self, game):
        new_status = game.status()
        self.games_view.item(
            game.short_name,
            values=(game.name, game.status_str()),
            image=self._icons[new_status])
        if new_status == GameStatus.COLLECTED:
            self.progress_bar.step(1)

    def new_status(self, status_text):
        self.status.config(text=status_text)
        logger.info(status_text)

    def _add_buttons(self, frm):
        buttons = ttk.Frame(frm, padding=10)
        buttons.grid(row=3, column=1, columnspan=2)
        buttons.grid()
        for i, _ in enumerate(self.tetsudoru.strings):
            plural = "es" if 0 < i else ""
            ttk.Button(
                buttons,
                text=f"Tetsudoru ({i + 1} guess{plural})",
                command=(lambda i: lambda: self.tetsudoru.to_clipboard(i + 1))(i)
            ).grid(column=i, row=1)
            quit_button = ttk.Button(buttons, text="Quit", command=self.root.destroy)
            quit_button .grid(column=0, row=2, columnspan=3, sticky='nesw')
            self.quit_button = ttk.Button(
                buttons,
                text="Clean up and finish",
                state=DISABLED)
            self.quit_button.grid(column=0, row=3, columnspan=3, sticky='nesw')
    def _add_clipboard_frame(self, frm):
        clipboard_frame = ttk.Frame(frm, borderwidth=1, width=400, padding=10)
        clipboard_frame.grid(column=0, row=0, rowspan=4, sticky='nesw')
        clipboard_frame.grid()
        ttk.Label(clipboard_frame, text="Clipboard", anchor="w",).grid(column=0, row=0)
        self.clipboard_display = Text(clipboard_frame, width=50, height=30)
        self.clipboard_display.config(state=DISABLED)
        self.clipboard_display.grid(column=0, row=1, sticky='nesw')
    def _add_progress_bar(self, frm):
        ttk.Label(frm, text='Collection progress', width=1).grid(row=2, column=1, sticky='ew')
        self.progress_bar = ttk.Progressbar(
            frm,
            maximum=sum(1 for i in self.col.collection if i.eligiblep())
        )
        self.progress_bar.grid(row=2, column=2, sticky='nesw')
        self.progress_bar.step(sum(1 for i in self.col.collection
                           if i.status() == GameStatus.COLLECTED))
    def _add_games_view(self, frm):
        self.games_view = ttk.Treeview(
            frm,
            padding=10,
            columns=('name', 'status'),
            height=len(self.col.collection))
        self.games_view.column('#0', width=50)
        self.games_view.heading('name', text='Name')
        self.games_view.column('name', width=150)
        self.games_view.heading('status', text='Status')
        self.games_view.column('status', width=300)
        self.games_view.grid(column=1, row=1, columnspan=2)
        for game in self.col.collection:
            self.games_view.insert(
                '', 'end', game.short_name,
                values=(game.name, game.status_str()),
                image=self._icons[game.status()])
    def _add_status_bar(self, frm):
        ttk.Separator(frm, orient=HORIZONTAL).grid(row=1, column=0, columnspan=3, sticky='ew')
        self.status = ttk.Label(frm, text='')
        self.status.grid(row=2, column=0, sticky='ew')
    def _mark_complete(self, _event):
        self.complete = True

    def exec(self):
        top_frame = ttk.Frame(self.root)
        top_frame.grid()
        main_widget_frame = ttk.Frame(top_frame, padding=10)
        main_widget_frame.grid(row=0, column=0)
        main_widget_frame.grid()
        ttk.Label(main_widget_frame, text=self._title()).grid(row=0, column=1, columnspan=2)
        self._add_games_view(main_widget_frame)
        self._add_buttons(main_widget_frame)
        # self._add_clipboard_frame(main_widget_frame)
        self._add_progress_bar(main_widget_frame)
        self._add_status_bar(top_frame)

        self.root.bind('q', lambda _: self.root.destroy())
        self.root.bind(ClipwaitThread.finish_event_name, self._mark_complete, add='+')

        if self.col.status() == CollectionStatus.COMPLETE:
            self.root.event_generate(ClipwaitThread.finish_event_name)
        else:
            self.clipwait_thread.start()
            self.new_status('Listening on clipboard...')
        self.root.mainloop()
        return CollectionStatus.COMPLETE if self.complete else self.col.status()



#!/usr/bin/env python3

import glob
from datetime import datetime
import os
import logging
import py7zr
import pytesseract
from PIL import Image
import re

from .collect_tk import Worker
from .res import Audio

logger = logging.getLogger(__name__)

yaml_fragment = """\
? ["{game}", "{date}"]
:
  dat: |
{text}
  timestamp: {timestamp}
"""

class Archiver:
    def __init__(self, config):
        self.config = config

    def game_object(self, filename):
        with open(filename) as f:
            dat = f.readlines()
        timestamp = datetime.fromtimestamp(
            os.stat(filename).st_mtime,
            tz=self.config.timezone
        ).isoformat()
        game, date = filename.split('-', 1)
        date = date[:-4]
        return yaml_fragment.format(
            game=game,
            date=date,
            timestamp=timestamp,
            text=''.join(' ' * (4 if l != '\n' else 0) + l for l in dat).rstrip())

    def strings_for_date(self, datestring):
        return ''.join(self.game_object(i) for i in glob.glob(f'*-{datestring}.txt'))

    def dump_to_file(self):
        outfile = self.config.yaml_archive_file()
        with open(outfile, 'a') as f:
            f.write(self.strings_for_date(self.config.date.strftime('%Y-%m-%d')))
        logger.info(f'Data sent to {outfile}.')

    def archive_to_7z(self):
        outfile = self.config.zip_archive_file()
        with py7zr.SevenZipFile(outfile, 'a') as archive:
            for i in (glob.glob(f'*-{self.config.date:%Y-%m-%d}.png')
                      + glob.glob(f'*-{self.config.date:%Y-%m-%d}.txt')):
                archive.write(i)
                os.remove(i)
        logger.info(f'Old items sent to {outfile}.')

    def celtix_cleanup(self):
        celtix_filename = f'celtix-{self.config.date:%Y-%m-%d}.png'
        if os.path.isfile(celtix_filename):
            with Image.open(celtix_filename) as img:
                img_text = pytesseract.image_to_string(img.crop((0, 0, 222, 74)))
                if puzzle_number := re.search('Celtix #([0-9]+)', img_text):
                    celtix_dest = os.path.expanduser(self.config.celtix_archive.format(
                        puznum=puzzle_number.group(1),
                        date=self.config.date))
                    os.rename(celtix_filename, celtix_dest)
                    logger.info(f'Moved Celtix result to {celtix_dest}.')
                else:
                    logger.warning(f'Cannot find Celtix puzzle number in {celtix_filename};')
                    logger.warning(f'Detected text was: {img_text!r}.')
        else:
            logger.info('No Celtix result found, skipping.')

class ArchiverWorker(Worker):
    event_name = "<<ArchiveComplete>>"
    def exec(self):
        if self.config.do_cleanup:
            cleanup = Archiver(self.config)
            cleanup.celtix_cleanup()
            cleanup.dump_to_file()
            cleanup.archive_to_7z()
    def hook(self):
        Audio.COMPLETE.play()
        self.gui.root.bind('<space>', lambda _: self.archive_and_quit())
        self.gui.quit_button.config(state='enabled', command=self.archive_and_quit)
    def archive_and_quit(self):
        self.config.do_cleanup = True
        self.gui.new_status("Cleaning up...")
        self.exec()
        self.gui.root.destroy()



import re
import os
import logging
from dataclasses import dataclass, field
import typing

from ..res import CHAR_TO_FILE

logger = logging.getLogger(__name__)

@dataclass(eq=True, frozen=True)
class GameRecord:
    game: str
    text: str | None
    image: bytes | None
    present: bool = True

    def __repr__(self):
        image = (f"<{len(self.image)} bytes>"
                 if self.image and len(self.image) > 100
                 else self.image)
        return (f"{self.__class__.__qualname__}(game={self.game!r}, "
                f"text={self.text!r}, image={image}, present={self.present})")

    def __str__(self):
        if self.present:
            return "<{class_name} {game}, text:{text}, image={image}>".format(
                class_name=self.__class__.__qualname__,
                game=f'{self.game!r}',
                text=f"'{self.text.splitlines()[0][:20]}...'" if self.text else 'None',
                image=f"<{len(self.image)} bytes>" if self.image else 'None',
            )
        else:
            game = self.game
            return f"{self.__class__.__qualname__}({game=}, text=None, image=None, present=False)"

### Parsing games
def parse_lines(linelist):
    blank_line = lambda line: line[0:1] in {'\n', chr(0xFE0F), ''}
    res = dict(caption=[], result_rows=[], rows=0, cols=0,)
    for line in linelist:
        if line.startswith("http"): # link lines, if they've escaped the filter earlier
            continue
        elif re.search('[A-Za-z0-9]', line): # result lines
            res['caption'].append(line + ('\n' if not line.endswith('\n') else ''))
        elif blank_line(line) and res['rows'] == 0: # blank lines
            continue
        elif blank_line(line):
            res['result_rows'].append('')
            res['rows'] += 1
        else:
            line = line.rstrip()
            if res['rows'] == 0:
                res['cols'] = sum(1 for i in line if i in CHAR_TO_FILE.keys())
            res['rows'] += 1
            res['result_rows'].append(line)
    while 0 < len(res['result_rows']) and res['result_rows'][-1] == '':
        res['result_rows'].pop()
    res['caption'] = ''.join(res['caption']).rstrip()
    res['success'] = not re.search('X/[0-9]|-/[0-9]|X&[0-9]/[0-9]', res['caption'])
    return res

MISSING_GAME = object()

@dataclass
class ParsedGame:
    source: typing.Any
    caption: str
    result_rows: list[str]
    rows: int
    cols: int
    success: bool
    image_data: None | bytes
    game_name: str | type(MISSING_GAME)
    extra_results: typing.Any = field(init=False, default=None)

    def __repr__(self):
        return (
            '{class_}(game={game!r}, rows={rows!r}, cols={cols!r}, '
            'caption={caption!r}, result_rows={result_rows!r}, image_data={image_data}, '
            'extra_results={extra_results!r}, success={success!r})'
        ).format(
            class_=self.__class__.__qualname__,
            game=self.game_name,
            caption=self.caption, rows=self.rows, cols=self.cols,
            result_rows=self.result_rows,
            success=self.success,
            image_data=(f"<{len(self.image_data)} bytes>"
                        if self.image_data and len(self.image_data) > 100
                        else repr(self.image_data)),
            extra_results=self.extra_results,)

    def __str__(self):
        short_caption = 'None' if self.caption is None else self.caption.splitlines()[0]
        return (
            '<{class_} {game} {rows}x{cols} caption={caption} {image_data} {success}>'
        ).format(
            class_=self.__class__.__qualname__,
            game=self.game_name,
            caption=repr((short_caption + '...') if '\n' in self.caption else short_caption),
            rows=self.rows, cols=self.cols,
            success='SUCCESS' if self.success else 'FAIL',
            image_data=(f'image_data:{len(self.image_data)} bytes'
                        if isinstance(self.image_data, bytes)
                        else f'image_data={self.image_data!r}'))

    @classmethod
    def missing_game(cls, game_name):
        return cls(
            game_name=MISSING_GAME,
            source=None,
            caption=game_name,
            result_rows=[], rows=0, cols=0, success=False,
            image_data=None)

def _id(x): return x # Used for defining reparse functions that don't do anything

class GameParser:
    def __init__(self, game_definitions):
        self.reparse_registry = dict()
        for i in game_definitions:
            if i.reparse_actions != []:
                self.reparse_registry[i.code] = self._make_reparse_function(i)

    def _make_reparse_function(self, game_definition):
        def reparse_function(game_to_reparse: ParsedGame) -> ParsedGame:
            for reparser in game_definition.reparse_actions:
                reparser.op(game_to_reparse)
            return game_to_reparse
        return reparse_function

    def _find_reparse(self, raw_parse):
        reparse_target = self.reparse_registry.get(raw_parse.game_name, _id)
        logger.info(f'Using {reparse_target} for {raw_parse.source}')
        return reparse_target(raw_parse)

    def load(self, linelist, source, image_data=None):
        res = parse_lines(linelist) | {"source": source, "game_name": source.short_name}
        maybe_image_file = (re.sub('txt$', 'png', source)
                            if isinstance(source, str)
                            else None)
        if isinstance(image_data, bytes):
            res['image_data'] = image_data
        elif maybe_image_file is not None and os.path.isfile(maybe_image_file):
            res['image_data'] = maybe_image_file
        else:
            res['image_data'] = None
        return self._find_reparse(ParsedGame(**res))

    def load_string(self, string, cfg):
        # String shall be identified first before parsing
        for game in cfg.collection.collection:
            if game.matches(string):
                return self.load(string.splitlines(), game)

    def load_record(self, db_result):
        if game_text := db_result.text:
            return self._find_reparse(
                ParsedGame(**parse_lines(game_text.splitlines()),
                           image_data=db_result.image,
                           game_name=db_result.game,
                           source=db_result))
        else:
            logger.warn(f'No data for {db_result}; using missing_game')
            return ParsedGame.missing_game(db_result.game)



import textwrap
import math
import logging
import re
from types import SimpleNamespace
from enum import Enum

logger = logging.getLogger(__name__)

class GameParseError(Exception):
    pass

class SuccessMetric(Enum):
    ALWAYS_SUCCEED = "always_succeed"
    LAST_ROW_UNIFORM = "last_row_uniform"
    NOT_SCORE_ZERO = "not_score_zero"

class Reparser:
    config_name = None
    @classmethod
    def from_dict(cls, spec):
        spec.pop("action")
        return cls(**spec)

    def op(self, game):
        pass

class AlternateSuccessMetric(Reparser):
    config_name = "success_metric"
    def __repr__(self):
        return f'{type(self).__qualname__}(metric={self.style.value!r})'
    def __init__(self, metric):
        self.style = SuccessMetric(metric)

    def op(self, game):
        if self.style == SuccessMetric.ALWAYS_SUCCEED:
            game.success = True
        elif self.style == SuccessMetric.NOT_SCORE_ZERO:
            game.success = game.extra_results.score != 0
        elif self.style == SuccessMetric.LAST_ROW_UNIFORM:
            last_row = game.result_rows[-1]
            game.success = last_row == last_row[0] * len(last_row)

class IntegerScoreToSquares(Reparser):
    config_name = "integer_score_to_squares"
    def __repr__(self):
        return f'{type(self).__qualname__}(columns={self.cols!r})'

    def __init__(self, columns=8):
        self.cols = columns

    def op(self, game):
        game.cols = self.cols
        score = game.extra_results.score
        game.rows = math.ceil(score / self.cols)
        cells = game.rows * self.cols
        game.result_rows = textwrap.wrap(
            (chr(0x1F7E5) * score).ljust(cells, chr(0x2B1C)),
            self.cols)

def _parse_types(thing):
    if thing == 'int': return int
    elif thing == 'float': return float
    elif thing == 'str': return str
    else:
        logger.warn(f'Unknown item type {thing} given, defaulting to str')
        return str

class ParseAndReformatCaption(Reparser):
    config_name = "reformat_caption"
    def __repr__(self):
        return (f'{type(self).__qualname__}(regexp={self.regexp!r}, '
                f'capture_groups={self.capture_groups!r}, '
                f'format_string={self.format_string!r})')

    def __init__(self, regexp, capture_groups, format_string=None, types=None):
        self.format_string = format_string
        self.regexp = re.compile(regexp)
        self.capture_groups = capture_groups
        self.types = ([str] * len(self.capture_groups)) if types is None else [_parse_types(i) for i in types]
    @classmethod
    def from_dict(cls, spec):
        return cls(
            regexp=spec.get("caption_regexp"),
            capture_groups=spec["capture"],
            format_string=spec.get("format_string"),
            types=spec.get('types'),)

    def op(self, game):
        default_string = f"{game.game_name.title()} {{puzzle_number}}"
        if dat := re.search(self.regexp, game.caption):
            captures = {i: dat.group(n + 1) for n, i in enumerate(self.capture_groups)}
        else:
            raise GameParseError
        for t, (k, v) in zip(self.types, captures.items()):
            captures[k] = t(v)
        game.extra_results = SimpleNamespace(**captures)
        game.caption = (self.format_string or default_string).format(**captures)

class Hexcodle(Reparser):
    config_name = "hexcodle"
    def op(self, game):
        if dat := re.search(
                '^I got Hexcodle #([0-9]+) in ([0-9])! Score: ([0-9]+)%', game.caption):
            game.caption = f'Hexcodle {dat.group(1)} {dat.group(2)}/5 @ {dat.group(3)}%'
            game.success = True
        elif dat := re.search(
                '^I didn\'t get Hexcodle #([0-9]+) :\\( Score: ([0-9]+)%', game.caption):
            game.caption = f'Hexcodle {dat.group(1)} X/5 @ {dat.group(2)}%'
            game.success = False
        else:
            raise GameParseError
        return game



import configparser
import tempfile
import subprocess
import platformdirs

def install_desktop_file():
    desktop_file = configparser.RawConfigParser()
    desktop_file.optionxform = lambda option: option
    desktop_file['Desktop Entry'] = {
        'Type': 'Application',
        'Exec': 'autowordle',
        'Name': 'Autowordle',
        'GenericName': 'Wordle clone collector',
        'Categories': 'Game',
    }
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = tmpdir/'autowordle.desktop'
        with open(filename, 'w') as f:
            desktop_file.write(f)
        subprocess.run([
            'desktop-file-install',
            '--dir=' + platformdirs.user_data_dir('applications'),
            filename
        ])

if __name__ == "__main__":
    install_desktop_file()



import logging
import os
import sys
from datetime import datetime
import yaml
import toml
from types import SimpleNamespace
import pytz
import subprocess
import argparse
import glob
import platformdirs

from .collect import GameCollection, CollectionStatus
from .game.detect import GameDefinition
from . import db
from . import collect_tk
from . import mkimg
from . import archiver
from . import desktop
from . import res

logger = logging.getLogger(__name__)

def load_definitions(*, more_definitions_file=None, selected_games=None):
    return [GameDefinition(dict(short_name=k) | v)
            for (k, v)
            in (res.definitions()
                | (toml.load(more_definitions_file) if more_definitions_file else dict())
                ).items()
            if (k in selected_games if selected_games is not None else True)]

def make_outimg_config(dat):
    return SimpleNamespace(
        width=dat['width'],
        game_separation=dat['game_separation'],
        title_size=dat['title']['font_size'],
        subtitle_size=dat['title']['subtitle_font_size'],
        font=dat['title']['font'])

def make_outimg_font(dat):
    return SimpleNamespace(family=dat['content_font'], size=dat['content_font_size'],)

def make_timeline_spec(dat):
    return SimpleNamespace(
        left_edge=dat['left_edge'],
        border_width=dat['border_width'],
        height_per_bucket=dat['height_per_bucket'],
        width_per_second=dat['width_per_second'],
        font=dat['font'],
        font_size=dat['font_size'],
        colour=dat['colour'])

class Config:
    program_name = 'Autowordle'
    default_config_filename = platformdirs.user_config_path(
        'autowordle', ensure_exists=True) / 'games.yaml'
    home_dir = platformdirs.user_data_dir('autowordle', ensure_exists=True)
    def __repr__(self):
        return '<{class_} {file_loc} {date:%Y-%m-%d}>'.format(
            class_=self.__class__.__qualname__,
            file_loc=self.location,
            date=self.date)
    def __init__(self, args):
        self.location = args.config
        with open(self.location) as f:
            self.dat = yaml.load(f, yaml.SafeLoader)
        self.do_cleanup = args.cleanup
        self.do_collect = args.collect
        self.check_already_collected = args.check_collected
        self.install_desktop_file = args.install_desktop_file
        self.timezone = pytz.timezone(self.dat['timezone'])
        self.date = datetime.now(tz=self.timezone)
        self.db = db.Db(self)
        self._outfile = os.path.expanduser(self.dat['output']['filename'])
        self.yaml_archive = self.dat['archives']['yaml']
        self.zip_archive = self.dat['archives']['zip']
        self.celtix_archive = self.dat['archives']['celtix']
        self.outimg_config = make_outimg_config(self.dat['output'])
        self.outimg_font = make_outimg_font(self.dat['output'])
        self.timeline = make_timeline_spec(self.dat['output']['timeline'])
        self.parsed_games = load_definitions(selected_games=self.dat['games'])
        self.collection = GameCollection(self.parsed_games, self.db)

    @classmethod
    def from_cmdline(cls):
        parser = argparse.ArgumentParser(
            description='Store and construct summary images for Wordle clones')
        for opt, desc in [('--cleanup', ('Clean up after image construction'
                                         ' (Overrideable in GUI)')),
                          ('--check-collected',
                           ("Exit if today's game is already collected"
                            " and collection is not in progress")),
                          ('--collect', 'Collect games')]:
            parser.add_argument(
                opt,
                action=argparse.BooleanOptionalAction,
                help=desc,
                default=True)
        parser.add_argument(
            '--install-desktop-file', help='Install a .desktop file, then exit',
            action='store_true')
        parser.add_argument(
            '--config',
            metavar='FILENAME',
            help='Configuration file location',
            default=cls.default_config_filename)
        return cls(parser.parse_args())

    @classmethod
    def from_args(cls, config_file=None, *,
                  collect=True, check_collected=True, cleanup=True,
                  install_desktop_file=False):
        return cls(SimpleNamespace(
            config=config_file or cls.default_config_filename,
            cleanup=cleanup,
            collect=collect,
            check_collected=check_collected,
            install_desktop_file=install_desktop_file))

    def autowordle_outfile(self):
        return self.date.strftime(os.path.expanduser(self._outfile))

    def yaml_archive_file(self):
        return self.date.strftime(os.path.expanduser(self.yaml_archive))

    def zip_archive_file(self):
        return self.date.strftime(os.path.expanduser(self.zip_archive))

    def window_titles(self):
        return {x.window_title for x in self.parsed_games}

class EarlyExit(Exception):
    def __init__(self, exit_code=1):
        self.exit_code = exit_code
    def exit(self):
        sys.exit(self.exit_code)

def check_already_collected(cfg):
    if (os.path.isfile(cfg.autowordle_outfile())
        and cfg.check_already_collected
        and len(glob.glob(cfg.home_dir + '*.txt')) == 0):
        status = subprocess.run([
            'yad',
            "--button=Copy today's image!gtk-ok",
            '--button=Close!gtk-cancel',
            '--image=dialog-warning',
            '--window-icon=dialog-warning',
            '--text', f"Today ({cfg.date:%Y-%m-%d})'s games are already collected.",
            '--title', cfg.program_name
        ])
        if status.returncode == 0:
            mkimg.send_result_to_clipboard(cfg)
        raise EarlyExit(1)

def top():
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    cfg = Config.from_cmdline()
    os.chdir(cfg.home_dir)

    try:
        if cfg.install_desktop_file:
            desktop.install_desktop_file()
            raise EarlyExit(0)
        check_already_collected(cfg)
        workers = [
            (mkimg.MkimgWorker(cfg), collect_tk.ClipwaitThread.finish_event_name),
            (archiver.ArchiverWorker(cfg), mkimg.MkimgWorker.event_name)]
        if cfg.do_collect:
            gui = collect_tk.GUI(cfg)
            cfg.do_cleanup = False # Don't clean up by default in GUI
            for worker, trigger in workers:
                gui.add_worker(worker, trigger)
            match gui.exec():
                case CollectionStatus.COMPLETE:
                    pass
                case CollectionStatus.COMPLETE_FOR_NOW:
                    logger.info('Collection complete for now')
                case CollectionStatus.INCOMPLETE:
                    logger.info('Collection aborted')
                    raise EarlyExit(1)
        else:
            for worker, _ in workers:
                worker.exec()
    except EarlyExit as e:
        e.exit()



from types import SimpleNamespace
from importlib import resources
from enum import Enum
import subprocess
import os
import toml

def colr_res(colour):
    return f"canvas:{colour}"

def file_res(filename):
    return resources.files(__package__)/filename

def definitions():
    return toml.load(file_res('game_defs.toml'))

class Audio(Enum):
    COMPLETE = "/usr/share/sounds/freedesktop/stereo/complete.oga"
    READY_FOR_TEXT = "~/.local/share/sounds/ready.mp3"
    def play(self):
        subprocess.Popen(["paplay", os.path.expanduser(self.value)])

colours = SimpleNamespace(
    theme1="#AF2303", theme2="#FFC800",
    border="#222222",
    black="#31373D",
    white="#E6E7E8",
    red="#DD2E44",
    blue="#55ACEE",
    orange="#F4900C",
    yellow="#FDCB58",
    green="#78B159",
    purple="#AA8ED6",
    brown="#C1694F",
)

CHAR_TO_FILE = {
    # Generic
    chr(0x2B1B):  colr_res(colours.black),
    chr(0x25A1):  colr_res(colours.white), # old-style
    chr(0x2B1C):  colr_res(colours.white), # new-style
    chr(0x1F7E5): colr_res(colours.red),
    chr(0x1F7E6): colr_res(colours.blue),
    chr(0x1F7E7): colr_res(colours.orange),
    chr(0x1F7E8): colr_res(colours.yellow),
    chr(0x1F7E9): colr_res(colours.green),
    chr(0x1F7EA): colr_res(colours.purple),
    chr(0x1F7EB): colr_res(colours.brown),

    # Cell Tower
    chr(0x2B07): file_res("down.png"),
    chr(0x2B06): file_res("up.png"),
    chr(0x27A1): file_res("right.png"),
    chr(0x2B05): file_res("left.png"),
    chr(0x23F9): colr_res(colours.blue),

    # Hexcodle
    chr(0x23EB):  file_res("dup.png"),
    chr(0x1F53C): file_res("up.png"),
    chr(0x2705):  colr_res(colours.green),
    chr(0x1F53D): file_res("down.png"),
    chr(0x23EC):  file_res("ddown.png"),

    # Kotoba de Asobou
    chr(0x2194):  file_res("lr.png"),
    chr(0x2195):  file_res("ud.png"),
    chr(0x1F7E2): colr_res(colours.orange),
}



from enum import Enum
import re
import io
import os
from PIL import Image
from datetime import datetime
import subprocess
import logging

from .res import Audio
from .game import Detector, DetectionType

logger = logging.getLogger(__name__)

class GameAccepts(Enum):
    NONE = "none"
    TEXT = "text"
    IMAGE = "image"

class GameStatus(Enum):
    OPTIONAL = 'optional'
    PARTIALLY_COLLECTED = 'partially_collected'
    COLLECTED = 'collected'
    UNCOLLECTED = 'uncollected'
    UNELIGIBLE = 'uneligible'
    def as_str(self):
        if self == GameStatus.OPTIONAL:
            return 'Optional'
        elif self == GameStatus.PARTIALLY_COLLECTED:
            return 'Awaiting text'
        elif self == GameStatus.COLLECTED:
            return 'Collected'
        elif self == GameStatus.UNCOLLECTED:
            return 'Uncollected'
        elif self == GameStatus.UNELIGIBLE:
            return 'Ineligible'

class Game:
    def __init__(self, spec, db):
        self.date = datetime.today().strftime("%Y-%m-%d")
        self.gui = None
        self.db = db
        self.detector = Detector(spec)
        self.name = spec.name
        self.display_row = spec.display_row
        self.refresh_hour = spec.refresh_hour
        self.short_name = self.detector.short_name
        self.out_filename = f'{self.detector.short_name}-{self.date}'
        self.result = self.db.load(self.date, self.short_name)

    def log(self, text):
        if self.gui is not None:
            self.gui.new_status(text)
        else:
            logger.info(text)

    def __repr__(self):
        return "<{class_name} {game_name}{collectedp}>".format(
            class_name=type(self).__qualname__,
            game_name=str(self.name),
            collectedp=" collected" if self.collectedp() else "",
        )

    def eligiblep(self):
        return self.refresh_hour <= datetime.now().hour

    def collectedp(self):
        if self.detector.detection == DetectionType.TEXT:
            return self.db.peek(self.date, self.short_name).text == 1
        elif self.detector.detection == DetectionType.IMAGE:
            dat = self.db.peek(self.date, self.short_name)
            return dat.text == 1 and dat.image == 1

    def status(self):
        dat = self.db.peek(self.date, self.short_name)
        has_image_but_no_text = dat.image == 1 and dat.text == 0
        if self.collectedp():
            return GameStatus.COLLECTED
        elif not self.eligiblep():
            return GameStatus.UNELIGIBLE
        elif has_image_but_no_text:
            return GameStatus.PARTIALLY_COLLECTED
        else:
            return GameStatus.UNCOLLECTED
        # else:
        #     return GameStatus.OPTIONAL

    def status_str(self):
        if self.result.text is not None:
            return self.result.text.splitlines()[self.display_row]
        else:
            return self.status().as_str()

    def save_text_to_file(self, thing):
        with open(self.out_filename + '.txt', 'w') as f:
            f.write("\n".join(line for line in thing.splitlines()
                              if not line.startswith("http")))
            f.write("\n")

    def save_to_db(self, thing):
        if isinstance(thing, str):
            r = self.db.save(self.date, self.short_name,
                         "\n".join(line for line in thing.splitlines()
                                   if not line.startswith("http")))
        elif isinstance(thing, bytes): # image
            r = self.db.save(self.date, self.short_name, thing)
        else:
            raise ValueError(f"Don't know what to do with {thing} on saving")
        return r

    def wants(self) -> GameAccepts:
        self.result = self.db.load(self.date, self.short_name)
        if self.collectedp() or not self.eligiblep():
            return GameAccepts.NONE
        elif self.detector.detection == DetectionType.TEXT or (
                self.detector.detection == DetectionType.IMAGE
                and self.result.image is not None):
            return GameAccepts.TEXT
        elif self.detector.detection == DetectionType.IMAGE and self.result.image is None:
            return GameAccepts.IMAGE
        else:
            logger.error('Unreachable want!')

    def collect_if_match(self, thing):
        if self.wants() == GameAccepts.NONE:
            return None
        elif (self.wants() == GameAccepts.TEXT
              and isinstance(thing, str)
              and self.detector.matches(thing)):
            self.save_text_to_file(thing)
            self.result = self.save_to_db(thing)
            self.log(f'Collected text of {self.name}')
            Audio.COMPLETE.play()
        elif (self.wants() == GameAccepts.IMAGE
              and isinstance(thing, Image.Image)
              and self.detector.matches(thing)):
            image_filename = self.out_filename + ".png"
            thing.save(image_filename)
            subprocess.run(['pngcrush', '-ow', image_filename])
            with open(image_filename, 'rb') as f:
                self.result = self.save_to_db(f.read())
            Audio.READY_FOR_TEXT.play()
            self.log(f'Collected image of {self.name}')
        else:
            # logger.error(f'Item {self} does not match')
            # logger.error(f'{self.wants()=}')
            # logger.error(f'{self.matches(thing)=}')
            # logger.error(f'{type(thing)=}')
            return None
        return self

class CollectionStatus(Enum):
    INCOMPLETE = 'incomplete'
    COMPLETE_FOR_NOW = 'complete-for-now'
    COMPLETE = 'complete'

class GameCollection:
    def __init__(self, collection, db):
        self.collection = [Game(i, db) for i in collection]

    def __repr__(self):
        return f"<{type(self).__qualname__}, {len(self.collection)} entries>"

    def inject_gui(self, gui):
        for i in self.collection:
            i.gui = gui

    def eligible(self):
        return [x for x in self.collection if x.eligiblep()]

    def status(self):
        if all(x.collectedp() for x in self.collection):
            return CollectionStatus.COMPLETE
        elif all(x.collectedp() for x in self.eligible()):
            return CollectionStatus.COMPLETE_FOR_NOW
        else:
            return CollectionStatus.INCOMPLETE

    def get(self, game_name):
        return next(x for x in self.collection if x.name == game_name)

    def find_matching_game(self, thing):
        for i in self.eligible():
            if res := i.collect_if_match(thing):
                return res
        return None

def play_sound(path):
    subprocess.Popen(["paplay", os.path.expanduser(path)])

def clipboard():
    if re.search(
            "image/",
            subprocess.run(
                ["xclip", "-sel", "c", "-target", "TARGETS", "-out"],
                capture_output=True).stdout.decode("UTF-8")):
        return Image.open(
            io.BytesIO(
                subprocess.run(
                    ["xclip", "-sel", "c", "-target", "image/png", "-out"],
                    capture_output=True).stdout))
    else:
        return subprocess.run(
            ["xclip", "-sel", "c", "-out"],
            capture_output=True).stdout.decode("UTF-8")



from . import prog
from . import mkimg
from . import collect
from . import collect_tk
from . import timer
from . import game
from .prog import Config

__all__ = ['prog', 'mkimg', 'collect', 'game', 'collect_tk', 'timer', 'Config']



import pytesseract
import re
from enum import Enum
from PIL import Image
from dataclasses import dataclass, field
import logging
import typing

from .reparse import Reparser

logger = logging.getLogger(__name__)

class DetectionType(Enum):
    TEXT = "text"
    IMAGE = "image"

@dataclass
class GameDefinition:
    code: str
    name: str
    target: re.Pattern
    text_target: re.Pattern | None
    window_title: str
    detection: DetectionType = DetectionType.TEXT
    refresh_hour: int = 0
    required: bool = True
    display_row: int = 0
    crop: typing.Optional[typing.Tuple[int, int, int, int]] = None
    reparse_actions: list[Reparser] = field(default_factory=list)

    def __init__(self, spec):
        if 'name' not in spec:
            raise ValueError('Must provide at least the name of the game.')
        dat = {
            "detection": DetectionType.TEXT,
            "refresh_hour": 0,
            "short_name": spec['name'].lower(),
            "name": None,
            "required": True,
            "target": "^" + spec["name"],
            "text_target": None,
            "display_row": 0,
            "crop": None,
            "reparse_actions": [],
            "window_title": spec['name'],
        } | spec
        self.window_title = dat['window_title']
        self.name = dat["name"]
        self.required = dat["required"]
        self.detection = DetectionType(dat["detection"])
        self.refresh_hour = dat["refresh_hour"]
        self.display_row = dat["display_row"]
        self.code = dat["short_name"]
        self.target = re.compile(dat['target'])
        self.text_target = re.compile(dat['text_target'] or self.target)
        match dat['crop']:
            case [x, y, w, h]: self.crop = (x, y, w, h)
            case None | []: self.crop = None
            case _:
                raise ValueError('Invalid crop spec: '
                                 'must be missing or an array of four integers.')

        self.reparse_actions = list()
        subclasses = Reparser.__subclasses__()
        for i in dat["reparse_actions"]:
            try:
                self.reparse_actions.append(
                    next(c for c in subclasses
                         if c.config_name == i["action"]).from_dict(i)
                )
            except StopIteration as e:
                raise ValueError(f'Unknown reparser {i["action"]}') from e
        if not re.fullmatch('^[a-z]+$', self.code):
            raise ValueError("A game code can only consist of lowercase letters.")
        if not (self.name and self.target):
            raise ValueError("name and target must be defined.")

class Detector:
    def __init__(self, spec):
        self.name = spec.name
        self.short_name = spec.code
        self.detection = spec.detection
        self.target = spec.target
        self.text_target = spec.text_target
        self.crop = spec.crop

    def __str__(self):
        return "<{class_} {short_name} ({name}) {target}>".format(
            class_=self.__class__.__qualname__,
            short_name=self.short_name,
            name=self.name,
            target=self.target,
        )

    def matches(self, thing):
        if isinstance(thing, str) and self.detection == DetectionType.TEXT:
            return self.target.search(thing)
        elif isinstance(thing, str) and self.detection == DetectionType.IMAGE:
            return self.text_target.search(thing)
        elif isinstance(thing, Image.Image) and self.crop is not None:
            return self.target.search(
                pytesseract.image_to_string(thing.crop(self.crop)))
        elif isinstance(thing, Image.Image):
            return self.target.search(pytesseract.image_to_string(thing))
        else:
            raise ValueError(f"Don't know what to do with {thing} on matching")



import sqlite3
import logging
import yaml
import contextlib
from datetime import datetime
from dataclasses import dataclass
import os

from .game.parse import GameRecord

logger = logging.getLogger(__name__)

@dataclass(eq=True, frozen=True)
class RecordPresence:
    text: bool
    image: bool
    exists: bool

class Db:
    def __enter__(self):
        self.con = sqlite3.connect(self.db_location)
        return self.con

    def __exit__(self, _exc_typ, _exc_val, _exc_tb):
        self.con.close()

    def __init__(self, cfg):
        self.db_location = os.path.expanduser(cfg.dat['archives']['db'])
        self.tz = cfg.timezone
        with self as db:
            db.execute("""CREATE TABLE IF NOT EXISTS dat (
                            id         INTEGER,
                            date       DATE     NOT NULL
                                       DEFAULT (DATE('now') ),
                            game       TEXT     NOT NULL,
                            timestamp  DATETIME NOT NULL
                                       DEFAULT (DATETIME('now') ),
                            text_data  TEXT,
                            image_data BLOB,
                            UNIQUE (date, game),
                            PRIMARY KEY (id AUTOINCREMENT));""")

    insert_text_query = (
        'INSERT INTO dat (date, game, timestamp, text_data) '
        'VALUES (:date, :game, :timestamp, :data) '
        'ON CONFLICT (date, game) DO UPDATE SET text_data = excluded.text_data '
        'RETURNING game, text_data, image_data'
    )
    insert_binary_query = (
        'INSERT INTO dat (date, game, timestamp, image_data) '
        'VALUES (:date, :game, :timestamp, :data) '
        'ON CONFLICT (date, game) DO UPDATE SET image_data = excluded.image_data '
        'RETURNING game, text_data, image_data'
    )
    insert_none_query = (
        'INSERT into dat (date, game, timestamp) '
        'VALUES (:date, :game, :timestamp) '
        'RETURNING game, text_data, image_data'
    )
    def save(self, date, game, thing):
        with (self as db, contextlib.closing(db.cursor()) as cur, db):
            info = {
                "date": date,
                "game": game,
                "timestamp": datetime.now(tz=self.tz).isoformat(),
                "data": thing,
            }
            try:
                if isinstance(thing, str):
                    logger.debug("Inserting a string to the database")
                    r = cur.execute(self.insert_text_query, info)
                elif isinstance(thing, bytes):
                    logger.debug("Inserting bytes to the database")
                    r = cur.execute(self.insert_binary_query, info)
                elif thing is None:
                    logger.info("Inserting a blank line to the database")
                    r = cur.execute(self.insert_none_query, info)
                else:
                    logger.error(f"Don't know what to do with {thing}")
                    raise ValueError(f"Don't know what to do with {thing}")
                (game, text, image) = r.fetchone()
            except sqlite3.IntegrityError:
                logger.warning('Cannot insert game - may have already been inserted')
            else:
                return GameRecord(game, text, image)

    load_query = ("SELECT date, game, text_data, image_data FROM dat "
                  "WHERE date = ? AND game = ?")
    def load(self, date, game):
        with (self as db, contextlib.closing(db.cursor()) as cur):
            match cur.execute(self.load_query, (date, game)).fetchone():
                case (_, game, text, image):
                    return GameRecord(game, text, image)
                case None:
                    return GameRecord(game, None, None, False)

    peek_query = ("SELECT text_data IS NOT NULL AS has_text, "
                  "image_data IS NOT NULL AS has_image FROM dat "
                  "WHERE date = ? AND game = ?")
    def peek(self, date, game):
        with (self as db, contextlib.closing(db.cursor()) as cur):
            match cur.execute(self.peek_query, (date, game)).fetchone():
                case (text, image):
                    return RecordPresence(text == 1, image == 1, True)
                case None:
                    return RecordPresence(False, False, False)

    date_query = "SELECT game, text_data, image_data FROM dat WHERE date = ?"
    def load_date(self, date):
        with (self as db, contextlib.closing(db.cursor()) as cur):
            return [
                GameRecord(game, text, image)
                for (game, text, image) in cur.execute(self.date_query, (date,))
            ]

class TextArchiveLoader(yaml.SafeLoader):
    def construct_mapping(self, node):
        pairs = self.construct_pairs(node, deep=True)
        try:
            return dict(pairs)
        except TypeError:
            rv = {}
            for key, value in pairs:
                if isinstance(key, list):
                    key = tuple(key)
                    rv[key] = value
            return rv

TextArchiveLoader.add_constructor('tag:yaml.org,2002:map',
                                  TextArchiveLoader.construct_mapping)

def load_archives(filename):
    with open(filename) as f:
        return yaml.load(f, TextArchiveLoader)

def dump_to_db(filename, selected_date, db):
    for ((game, date), dat) in load_archives(filename).items():
        if selected_date == date:
            with contextlib.suppress(sqlite3.IntegrityError):
                db.save(dat['timestamp'], game, dat['dat'])
